local _, cdkp = ...

-- CORE --
cdkp.CORE_CONSTANTS = {};
cdkp.CORE_CONSTANTS.MINDKP = {
	5,	-- MC
	10	-- BWL
};

cdkp.CORE_CONSTANTS.MAP_IDS = {};
cdkp.CORE_CONSTANTS.MAP_IDS.BWL = 469;

function cdkp.Contains(array, item)
    for _, e in ipairs(array) do
        if item == e then
            return true
        end
    end
    return false
end

function cdkp.Remove(array, item)
    for i, e in ipairs(array) do
        if item == e then
            table.remove(array, i);
            break
        end
    end
end

function cdkp.GetRaidMembers()
    local out = {};
    for i=1,40 do
        local name, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(i);
        if name then
            table.insert(out, name);
        end
    end
    return out
end

-- bids are (spec [2|1], amount, ...)
function cdkp.MSOverOSSort(new_bid, old_bid)
    -- check spec, then bid amount
    if new_bid[1] < old_bid[1] then
        return true
    elseif new_bid[1] > old_bid[1] then
        return false
    else
        return new_bid[2] > old_bid[2]
    end
end

-- only used in 2:1 priority systems
function cdkp.BidWeight(bid)
    local spec, amt = unpack(bid);
    return amt / spec
end

-- bids are (spec [2|1], amount, ...)
function cdkp.TwoToOneSort(new_bid, old_bid)
    -- 2:1 means an OS bid has to be 2x an MS bid to tie, so compare bid weights
    return cdkp.BidWeight(new_bid) > cdkp.BidWeight(old_bid)
end

function cdkp.ApplyPriority(b1, b2)
    if cdkp.CONSTANTS.PRIORITY_TYPE == cdkp.CONSTANTS.MS_OVER_OS then
        return cdkp.MSOverOSSort(b1, b2);
    elseif cdkp.CONSTANTS.PRIORITY_TYPE == cdkp.CONSTANTS.TWO_TO_ONE then
        return cdkp.TwoToOneSort(b1, b2);
    end
end

function cdkp.InsertNewBid(bids, new_bid)
    -- find the insertion index then shift everything after downward

    -- default to inserting at the end, for an empty table this will be position 1
    local insert_at = #bids + 1;
    for i, old_bid in ipairs(bids) do
        if cdkp.ApplyPriority(new_bid, old_bid) then
            insert_at = i;
            break
        end
    end

    -- shift from the end so we don't overwrite values we have yet to shift
    for i=#bids,insert_at,-1 do
        bids[i+1] = bids[i]
    end
    bids[insert_at] = new_bid
end

function cdkp.StripRealm(character)
    local start = string.find(character, "-");
    if start then
        return character:sub(1, start - 1)
    else
        return character
    end
end

function cdkp.IsValidBid(character, amount)
    return amount >= cdkp.GetMinDKP() and cdkp.GetDKP(character) >= amount
end

function cdkp.GetDKP(character)
	--[[
    local num_guild_members = GetNumGuildMembers();
    for i=1,num_guild_members do
        local nameAndRealm, rank, rankIndex, _, class, zone, publicnote, officernote, online = GetGuildRosterInfo(i);
        if character == cdkp.StripRealm(nameAndRealm) then
            local _, _, dkp = string.find(officernote, "<(-?%d*)>");
            if dkp then
                dkp = tonumber(dkp);
            end
            if dkp then
                return dkp
            end
        end
    end
    return 0
	]]--
	return 500
end

function cdkp.CheckIfCharacterIsAllowed(character)
	local num_guild_members = GetNumGuildMembers();
    for i=1,num_guild_members do
        local nameAndRealm, rank, rankIndex, _, class, zone, publicnote, officernote, online = GetGuildRosterInfo(i);
        if character == cdkp.StripRealm(nameAndRealm) then
            if rank == "Offizier" then
				return true
			end
			return false
        end
    end
	return false
end

-- generates ~64 bit random ids
function cdkp.GenerateAuctionId()
    return string.gsub("xxxxxxxxxxxxxxxxxxxx", "[x]", function ()
        return string.format("%x", math.random(0, 0x9))
    end)
end

function cdkp.PrintableSpec(spec)
    if tonumber(spec) == 1 then
        return "MS"
    end
	if tonumber(spec) == 2 then
        return "OS"
	end 
	if tonumber(spec) == 3 then
		return "Trial"
	end 
	if tonumber(spec) == 4 then
		return "Twink"
    end
end

-- CORE --
function cdkp.GetMinDKP()

	local name, type, difficultyIndex, difficultyName, maxPlayers, dynamicDifficulty, isDynamic, instanceMapId, lfgID = GetInstanceInfo()
	
	-- BWL --
	if instanceMapId == cdkp.CORE_CONSTANTS.MAP_IDS.BWL then
		return cdkp.CORE_CONSTANTS.MINDKP[2]
	end
	
	return cdkp.CORE_CONSTANTS.MINDKP[1]
end